#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#ifdef __SSE4_2__
#include <nmmintrin.h>
#endif

/* some macro activated parameters */
/* #define WEIGHTED_SET */
#define POPCNT

#define SEEDFILE "seeds.txt"
#define SEEDSTRINGLENGTH 100
#define NBMAXSEEDS 2



#ifdef WEIGHTED_SET
#define WEIGHT(__a__) ((unsigned long long) binomial_weight[p_count] * (__a__))
#else
#define WEIGHT(__a__) (__a__)
#endif


/*
long long unsigned binomial_numbers[33] = {1,32,496,4960,35960,201376,906192,3365856,10518300,28048800,64512240,129024480,225792840,347373600,471435600,565722720,601080390,565722720,471435600,347373600,225792840,129024480,64512240,28048800,10518300,3365856,906192,201376,35960,4960,496,32,1};
*/

#ifdef WEIGHTED_SET
long long unsigned binomial_weight[33] = {257403837600,8043869925,518959350,51895935,7158060,1278225,284050,76475,24472,9177,3990,1995,1140,741,546,455,428,455,546,741,1140,1995,3990,9177,24472,76475,284050,1278225,7158060,51895935,518959350,8043869925,257403837600};
#endif

int main() {

  /* precompute binomial weights */
  /*
  int i;
  for (i=0;i<33;i++){
    fprintf(stdout,"%llu,//(%f)\n",257403837600/binomial_weight[i],(double)257403837600.0/(double)binomial_weight[i]);
  }
  return 0;
  */



  /* foreach seeds */
  FILE * f = fopen (SEEDFILE, "r");
  if (f == NULL) {
    perror ("Error opening file " SEEDFILE);
    return -1;
  }
  char seedstring[SEEDSTRINGLENGTH+1];
  while (fgets (seedstring, SEEDSTRINGLENGTH, f) != NULL) {

    unsigned nbseeds = 0;
    unsigned seeds[NBMAXSEEDS] = {0};
    unsigned seeds_len[NBMAXSEEDS] = {0};

    {
      char *pos;
      if ((pos=strchr(seedstring, '\n')) != NULL)
        *pos = '\0';
    }
    printf("\"%s\"",seedstring); fflush(NULL);

    /* (A) extract seeds and generate separated bits elements */
    char * tok = strtok(seedstring,";,");
    nbseeds = 0;
    while (tok) {
      /* seed length */
      seeds_len[nbseeds] = strlen(tok);
      /* seed pattern */
      seeds[nbseeds] = 0;
      {
        int i;
        for (i = 0; i < seeds_len[nbseeds]; i++) {
          if (tok[i] == '#')
            seeds[nbseeds] |= 1<<i;
        }
      }
      /* set seed length to its shift */
      seeds_len[nbseeds] = (sizeof(unsigned)*8) - seeds_len[nbseeds];
      nbseeds++;
      if (nbseeds > NBMAXSEEDS) {
        fprintf(stderr,"Error too many seeds\n");
        return 1;
      }
      tok = strtok(NULL, ";,");
    }


    /* (B) set stats counter then generate all alignments */
    long long unsigned x [33] = {0};
    long long unsigned y [33] = {0};
    long long unsigned z [33] = {0};
    long long unsigned x2[33] = {0};
    long long unsigned y2[33] = {0};
    long long unsigned z2[33] = {0};
    long long unsigned xz[33] = {0};
    long long unsigned yz[33] = {0};
    long long unsigned n [33] = {0};


    /* Quick SSE4.2 "128bits" hardcoded ... (no avx2 in the room ... arg!! ...)
     * (no current compiler does this trick ... so must be done manually)
     */
#ifdef  __SSE4_2__
    __m128i a_v   = _mm_set_epi32  (0xC0000000, 0x80000000, 0x40000000, 0x00000000);
    __m128i one_v = _mm_set1_epi32 (1);
    __m128i us_v[NBMAXSEEDS];
    {
      int s;
      for (s = 0; s < nbseeds; s++) {
        us_v[s] = _mm_set1_epi32(seeds[s]);
      }
    }



    int a_count;
    for (a_count = 0; a_count < 0x40000000; a_count++) {

      /* (B.1) compute coverage and multihit for all the 2^32 words */
      __m128i g_cov_w_v = _mm_setzero_si128();
      __m128i m_count_v = _mm_setzero_si128();
      int s;
      for (s = 0 ; s < nbseeds ; s++) {
        __m128i u_v = us_v[s];
        int i;
        for (i = 0; i <= seeds_len[s]; i++) {
          __m128i e_v              = _mm_cmpeq_epi32(_mm_and_si128(a_v, u_v), u_v);
          __m128i m_count_to_add_v = _mm_and_si128(e_v, one_v);
          __m128i g_cov_w_to_or_v  = _mm_and_si128(e_v, u_v);
          g_cov_w_v = _mm_or_si128(g_cov_w_v, g_cov_w_to_or_v);
          m_count_v = _mm_add_epi32(m_count_v, m_count_to_add_v);
          u_v       = _mm_slli_epi32(u_v, 1);
        }
      }

      /* (B.2) update stats counters for these 4 alignments : can be optimized too ...
       * (but both "clang 5.1"/"icc 14"/"gcc 4.8" do the job here so ...)
       */
      int i;
      unsigned p_count = _mm_popcnt_u32(a_count);

      for (i = 0; i < 4; i++) {

        p_count += (i & 1);
        unsigned m_count =                _mm_extract_epi32(m_count_v, i);
        unsigned g_cov   = _mm_popcnt_u32(_mm_extract_epi32(g_cov_w_v, i));
        unsigned long long m_count_add = WEIGHT(m_count);
        unsigned long long   g_cov_add = WEIGHT(g_cov);
        unsigned long long p_count_add = WEIGHT(p_count);

        /* do the sum for the sets that consider percentage to be higher than "p"/33 */

        /* partial sums can be done  for theses variables */
        x [p_count] += m_count_add;
        x2[p_count] += m_count_add * m_count;
        y [p_count] +=   g_cov_add;
        y2[p_count] +=   g_cov_add * g_cov;
        z [p_count] += p_count_add;
        z2[p_count] += p_count_add * p_count;
        xz[p_count] += m_count_add * p_count;
        yz[p_count] +=   g_cov_add * p_count;
        n [p_count] += WEIGHT(1);

        /* "p"/33 */
      }
      /* DEBUG STUFF ... MAY BE HELFULL ONE DAY ... */
      /*
      fprintf(stdout,"%x,%x,%x,%x\n",
             _mm_extract_epi32 (a_v, 0),
             _mm_extract_epi32 (a_v, 1),
             _mm_extract_epi32 (a_v, 2),
             _mm_extract_epi32 (a_v, 3));
      */
      /* next alignment */
      a_v = _mm_add_epi32 (a_v, one_v);
    }

#else /* #ifdef  __SSE4_2__ */

    /* Non-optimized code */
    unsigned a = 0;
    do {
      /* (B.1) compute coverage and multihit for all the 2^32 words */
      unsigned g_cov_w = 0;
      int m_count = 0;
      int s;
      for (s = 0; s < nbseeds; s++) {
        unsigned u = seeds[s];
        int i;
        for (i = 0; i <= seeds_len[s]; i++) {
          if ((a & u) == u) {
            m_count++;
            g_cov_w |= u;
          }
          u <<= 1;
        }
      }

#ifdef POPCNT
      int g_cov   = __builtin_popcount(g_cov_w);
      int p_count = __builtin_popcount(a);
#else
      int g_cov   = 0;
      int p_count = 0;
      {
        int i;
        for (i = 0; i < (sizeof(unsigned)*8); i++) {
          if (g_cov_w & (1<<i)) {
            g_cov++;
          }
          if (a & (1<<i)) {
            p_count++;
          }
        }
      }
#endif

      /* (B.2) update stats counters for this alignment */
      /* "p"/33 partial sums >> */
      x [p_count] += m_count_add;
      x2[p_count] += m_count_add * m_count;
      y [p_count] +=   g_cov_add;
      y2[p_count] +=   g_cov_add * g_cov;
      z [p_count] += p_count_add;
      z2[p_count] += p_count_add * p_count;
      xz[p_count] += m_count_add * p_count;
      yz[p_count] +=   g_cov_add * p_count;
      n [p_count] += WEIGHT(1);
      /* << "p"/33 */

      /* next alignment */
    } while (++a > 0);
#endif

    /* "p"/33 partial sums to full sums >> */
    {
      unsigned u;
      for(u = 32; u > 0 ; u--) {
        x [u-1] += x [u];
        x2[u-1] += x2[u];
        y [u-1] += y [u];
        y2[u-1] += y2[u];
        z [u-1] += z [u];
        z2[u-1] += z2[u];
        xz[u-1] += xz[u];
        yz[u-1] += yz[u];
        n [u-1] += n [u];
      }
    }
    /* << "p"/33 */

    /* computing and printing correlations */
    {
      unsigned p;
      for(p = 0; p < 32 /* "sigz == 0" if "p == 32" : one single sample only */; p++) {
        double covxz  =       (double) xz[p] * (double) n[p] - (double) x[p] * (double) z[p];
        double covyz  =       (double) yz[p] * (double) n[p] - (double) y[p] * (double) z[p];
        double sigx   = sqrt( (double) x2[p] * (double) n[p] - (double) x[p] * (double) x[p] );
        double sigy   = sqrt( (double) y2[p] * (double) n[p] - (double) y[p] * (double) y[p] );
        double sigz   = sqrt( (double) z2[p] * (double) n[p] - (double) z[p] * (double) z[p] );
        double corrxz = covxz / (sigx * sigz);
        double corryz = covyz / (sigy * sigz);
        printf("\t%g\t%g",corrxz,corryz);
      }
    }
    printf("\n"); fflush(NULL);
  } /* each seed */
  fclose(f);
  return 0;
}

